{-# LANGUAGE FlexibleInstances #-}

module Parser where

--- Parser a ------------------------------------------------------------------

type Parser a = [String] -> (Maybe a, [String])

anyToken :: Parser String
anyToken (tkn:tkns)          = (Just tkn, tkns)
anyToken []                  = (Nothing, [])

-- combinators ----------------------------------------------------------------

-- result -----------

result :: a -> Parser a
result x = \tkns -> (Just x, tkns)

-- andThen ----------

andThen  :: Parser a -> (a -> Parser b) -> Parser b
andThen p fn =
    \tkns ->
        case p tkns of
            (Nothing, _)    -> (Nothing, tkns)
            (Just a, rest)  -> (fn a) rest


oneToken :: Parser String
oneToken = anyToken `andThen` \tkn -> result tkn


twoTokens' :: Parser String
twoTokens' = anyToken `andThen` (\tkn1 -> anyToken)


twoTokens  :: Parser String
twoTokens  = anyToken `andThen`
                       (\t1 -> anyToken
                            `andThen`
                                (\t2 -> result (t1 ++ t2))
                       )

-- mapping and filtering -----------------------------------------------------

mapTo :: Parser a -> (a -> b) -> Parser b
mapTo p fn  = p `andThen` \a -> result (fn a)


anyTokenLength :: Parser Int
anyTokenLength = anyToken `mapTo` length


filterBy :: Parser a -> (a -> Bool) -> Parser a
filterBy p pred =
    \tkns ->
        case p tkns of
            (Nothing, rest) -> (Nothing, tkns)
            (Just x, rest)  -> if pred x then (Just x, rest)
                               else (Nothing, tkns)

token :: String -> Parser String
token tkn =  anyToken `filterBy` (== tkn)


open :: Parser String
open = undefined -- TODO


close :: Parser String
close = undefined -- TODO


word :: Parser String
word = undefined -- TODO


-- further combinators --------------------------------------------------------

-- multiple -----------

multiple :: Show a => Parser a -> Parser [a]
multiple p =
    \tkns ->
       let
          r@(mbP, pRest) = p tkns
       in
          case mbP of
              Nothing  -> (Just [], pRest)
              Just pR   ->
                 case multiple p pRest of
                    (Just results, rest)  -> (Just (pR:results), rest)
                    (Nothing, rest)       -> (Nothing, rest)


multipleAs :: Parser [String]
multipleAs = multiple (token "A")

threeAs :: (Maybe [String], [String])
threeAs = multipleAs ["A", "A", "A", "B", "C"]

noA :: (Maybe [String], [String])
noA = multipleAs ["B", "C"]

-- or ------------

orElse :: Parser a -> Parser a -> Parser a
orElse p q = undefined -- TODO


aOrB :: Parser String
aOrB = (token "A") `orElse` (token "B")


(parse1, rest1) =  aOrB ["A", "B", "C"]
(parse2, rest2) =  aOrB rest1
(noParseB, rest3) =  aOrB rest2


multipleAOrBs :: Parser [String]
multipleAOrBs = multiple ((token "A") `orElse` (token "B"))

(parse4, rest4) = multipleAOrBs ["A", "B", "B", "A", "C", "D"]


multipleAOrBsInBrackets :: Parser [String]
multipleAOrBsInBrackets = open `andThen`
                             (\_ -> multipleAOrBs `andThen`
                                  (\abs -> close `andThen`
                                      (\_ -> result abs)))

(parse5, rest5) = multipleAOrBsInBrackets ["(","A", "B", "B", "A", ")"]

-------------------------------------------------------------------------------
